/*
 * dburl
 * by Ole Tange & Hans Schou 2004
 *
 * Installed libraries:
 *   libpcre3-dev
 *   gcc
 *
 * Compile:
 *   gcc -Wall -lpcre -o dburl_c dburl.c
 *   gcc -Wall -lpcre -o dburl_c -DINCLUDE_COMMENTS=1 dburl.c
 *
 * Test:
 *   ./dburl_c 'pgsqls://example/' 'pgsqls://user:password@host:5432/database/SELECT * FROM mytable'
 *   ./dburl_c 'mysql://user:password@host:3306/database/table/column1,column2'
 *   ./dburl_c 'odbc+dsn:////table/column1,column2'
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <errno.h>
#include <pcre.h>

#include "dburl.h"

#define DEBUG 1

#define OVECCOUNT (50*3)

#define IDX_SCHEME	1
#define IDX_DSN		IDX_SCHEME+1
#define IDX_USER	IDX_DSN+1
#define IDX_PASSWORD	IDX_USER+1
#define IDX_HOST	IDX_PASSWORD+1
#define IDX_PORT	IDX_HOST+1
#define IDX_DATABASE	IDX_PORT+1
#define IDX_TABLE	IDX_DATABASE+1
#define IDX_COLUMN	IDX_TABLE+1
#define IDX_SQL		IDX_COLUMN+1

const char *schemeitems[] = {
	"null",
	"scheme",
	"dsn",
	"user",
	"password",
	"host",
	"port",
	"database",
	"table",
	"coloumn",
	"sql"
};

// If INCLUDE_COMMENTS comments would be included inside the regular expression
#ifdef INCLUDE_COMMENTS
#define cm(msg) "(?#\n " msg "\n)"
#else
#define cm(msg)
#endif

const char syntaxdescription[] =
"[sql:][scheme[<+>dsn]]://[[user[:[password]]@][host][:port][/[database][/[[table[/[column[,column...]*]]]|sql]]]"
;
const char dburlregularexpression[] =
"^"
	cm("Optional prefix 'sql:'")
	"(?:sql:)?"
	cm("Scheme: mysql|pgsql")
	"([-.a-z0-9]*)(?:[+]([-.a-z0-9]*))?"
	cm("Required: URL identifier")
	"://"
	cm("User name + password")
	"(?:"
		cm("Username")
		"([-a-z0-9_]+)"
		cm("Password")
		"(?::([^@]*))?@"
	")?"
	cm("Hostname")
	"("
		cm("localhost | example")
		"(?:[a-z0-9]+(?:-+[-a-z0-9]+)*)"
	"|"
		cm("Domain name with dot: example.com")
		"(?:(?:[a-z0-9]+(?:-+[a-z0-9]+)*\\.)?"
			"(?:[a-z0-9]+(?:-+[a-z0-9]+)*\\.)+(?:[a-z]{2,7})\\.?)"
	"|"
		cm("IPv4 number")
		"(?:(?:25[0-5]|2[0-4][0-9]|[0-1][0-9]{2}|[0-9][0-9]|[0-9])\\.){3}"
		"(?:25[0-5]|2[0-4][0-9]|[0-1][0-9]{2}|[0-9][0-9]|[0-9])"
	")?"
	cm("Port number: 3306|5432")
	"(?::(\\d{1,5}))?"
	cm("DB, table, SQL")
	"(?:/"
		"(?:"
			cm("Database: joe|mydb, default $USER")
			"(?:([_a-z0-9]+)?"
				"(?:/"
					"(?:"
						cm("Table: mytable")
						"(?:([_a-z0-9]+)"
							cm("Columns: id, name")
                  					"(?:/"
		                     				"((?:[_a-z0-9]+)"
				                         		"(?:,[_a-z0-9]+)*"
							        ")?"
				                        ")?"
						")|("
							cm("SQL: SELECT id, name FROM mytable")
							"[^\\h]+\\h.+"
						")?"
					")?"
				")?"
			")?"
		")?"
	")?"
"$"
;

char *termstring(char *txt, int *ov, int idx, char *para_def) {
	char *tmp = NULL;

	/* if there is a match on this index... */
	if (ov[2*idx+1] > 0) {
		int length = ov[2*idx+1] - ov[2*idx];
		if ((tmp = malloc(length+1))) {
			strncpy(tmp, &txt[ov[2*idx]], length);
			tmp[length] = 0;
		}
	}
	//return tmp;
	return tmp ? tmp : para_def;
}

tdburl *dburlparse(char *dburltext) {
	tdburl *dburl = NULL;

	if (dburltext) {
		pcre *re;
		const char *error; 
		int erroffset;
		int ovector[OVECCOUNT];
		int rc;

		int i;

        	re = pcre_compile(
				dburlregularexpression,
				PCRE_CASELESS,
				&error,
				&erroffset,
				NULL);
		if (re) {
			rc = pcre_exec(
				re,
				NULL,
				dburltext,
				strlen(dburltext),
				0,
				0,
				ovector,
				OVECCOUNT);
			if (rc > 0) {
#ifdef DEBUG
				int m = 0;
				int color = -1;
				printf("Input: '%s'\n", dburltext);
				printf("        ");
				for (i = 1; i < rc; i++) { 
					while (m < ovector[2*i]) {
						if (color >= 0) { printf("\e[0m"); color = -1; }
						printf("_");
						++m;
					}
					while (m < ovector[2*i+1]) {
						if (color != i % 2) { printf("\e[3%dm", i % 2 ? 5 : 3); color = i % 2; }
						printf("%d", i % 10);
						++m;
					}
				}
				if (color >= 0) { printf("\e[0m"); color = -1; }
				printf("\n");
				for (i = 0; i < rc; i++) { 
					char *substring_start = dburltext + ovector[2*i]; 
					int substring_length = ovector[2*i+1] - ovector[2*i]; 
					printf("# %2d [%2d %2d]: %.*s\n", i, ovector[2*i], ovector[2*i+1], substring_length, substring_start); 
				} 
#endif
				dburl = malloc(sizeof(*dburl));
				memset(dburl, 0, sizeof(*dburl));
				/* Assign all possible elements a value */
				dburl->scheme	= termstring(dburltext, ovector, IDX_SCHEME,   "<one-scheme-found>");
				dburl->dsn	= termstring(dburltext, ovector, IDX_DSN,      NULL);
				dburl->user	= termstring(dburltext, ovector, IDX_USER,     "<user-name>");
				dburl->password	= termstring(dburltext, ovector, IDX_PASSWORD, "<none>");
				dburl->host	= termstring(dburltext, ovector, IDX_HOST,     "localhost");
				dburl->port	= termstring(dburltext, ovector, IDX_PORT,     "<scheme-default>");
				dburl->database = termstring(dburltext, ovector, IDX_DATABASE, "<user-name>");
				dburl->table	= termstring(dburltext, ovector, IDX_TABLE,    NULL);
				dburl->column	= termstring(dburltext, ovector, IDX_COLUMN,   NULL);
				dburl->sql	= termstring(dburltext, ovector, IDX_SQL,      NULL);
			}
			pcre_free(re);
		}
	}

	return dburl;
}

void fieldfree(char *p) {
	if (p) {
		free(p);
	}
}

tdburl *dburlfree(tdburl *dburl) {
	if (dburl) {
		fieldfree(dburl->scheme);
		fieldfree(dburl->dsn);
		fieldfree(dburl->user);
		fieldfree(dburl->password);
		fieldfree(dburl->host);
		fieldfree(dburl->port);
		fieldfree(dburl->database);
		fieldfree(dburl->table);
		fieldfree(dburl->column);
		fieldfree(dburl->sql);
		free(dburl);
		dburl = NULL;
	}
	return NULL;
}

int main(int argc, char **argv) {
	int i = 1;

	printf("\n%s\n", syntaxdescription);
	printf("\n%s\n\n", dburlregularexpression);
	while (i<argc) {
		tdburl *url = dburlparse(argv[i]);

		if (url) {
			printf("%2d. %-10s: %s\n", IDX_SCHEME,   schemeitems[IDX_SCHEME],   url->scheme);
			printf("%2d. %-10s: %s\n", IDX_DSN,      schemeitems[IDX_DSN],      url->dsn);
			printf("%2d. %-10s: %s\n", IDX_USER,     schemeitems[IDX_USER],     url->user);
			printf("%2d. %-10s: %s\n", IDX_PASSWORD, schemeitems[IDX_PASSWORD], url->password);
			printf("%2d. %-10s: %s\n", IDX_HOST,     schemeitems[IDX_HOST],     url->host);
			printf("%2d. %-10s: %s\n", IDX_PORT,     schemeitems[IDX_PORT],     url->port);  
			printf("%2d. %-10s: %s\n", IDX_DATABASE, schemeitems[IDX_DATABASE], url->database);
			printf("%2d. %-10s: %s\n", IDX_TABLE,    schemeitems[IDX_TABLE],    url->table);
			printf("%2d. %-10s: %s\n", IDX_COLUMN,   schemeitems[IDX_COLUMN],   url->column);
			printf("%2d. %-10s: %s\n", IDX_SQL,      schemeitems[IDX_SQL],      url->sql);
		} else {
			printf("ERROR parsing '%s'\n", argv[i]);
		}
		printf("\n");
			
		// BUG: Had disable, can not free() constant default like "<scheme>"
		//dburlfree(url);
		i++;
	}

	return 0;
}
